
from pyroute2.netlink.iproute import iproute
from pyroute2.netlink.ipdb import ipdb
from pyroute2.netlink.taskstats import taskstats

make_pep8_happy = iproute
make_pep8_happy = ipdb
make_pep8_happy = taskstats
